//
// Simple functions to work with RGBE textures (in .hdr files)
//

#include	<math.h>
#include	"libExt.h"
#include	"libTexture.h"
#include	"Data.h"
#include	"StringUtils.h"

static	const char * sign = "#?RADIANCE";	// every .hdr file must start with it

inline	void rgbe2float ( float * floatPtr, const byte * rgbePtr )
{
	if ( rgbePtr [3] ) 
	{  
		float	f = ldexp ( 1.0, rgbePtr [3] - (int)(128+8));
	
		floatPtr [0] = rgbePtr [0] * f;
		floatPtr [1] = rgbePtr [1] * f;
		floatPtr [2] = rgbePtr [2] * f;
	}
	else
		floatPtr [0] = 
		floatPtr [1] = 
		floatPtr [2] = 0.0f;
}

bool	parseHdrHeader ( Data * data, string& format, int& width, int& height, float& exposure, float& gamma )
{
	string		str;
	
	exposure = 1.0f;
	gamma    = 1.0f;
	
	if ( !data -> getString ( str, '\n' ) )
		return false;
		
	if ( str != sign )
		return false;
											// now read string from header
	while ( data -> getString ( str, '\n' ) )
		if ( str.empty () )					// end of header
			break;
		else
		{
			int	pos = str.find ( '=' );		// split to ...
			
			if ( pos == string :: npos )
				continue;
				
			string	name  = trim ( str.substr ( 0, pos ) );
			string	value = trim ( str.substr ( pos + 1 ) );
			
			if ( name == "FORMAT" )
				format = value;
			else
			if ( name == "EXPOSURE" )
				exposure = atof ( value.c_str () );
			else
			if ( name == "GAMMA" )
				gamma = atof ( value.c_str () );
		}
		
	if ( format != "32-bit_rle_rgbe" && format != "32-bit_rle_xyze" )		// illegal format
		return false;
		
											// now parse resolution
	if ( !data -> getString ( str, '\n' ) )
		return false;
		
	int	px = str.find ( 'X' );
	int	py = str.find ( 'Y' );
	
	if ( px == string :: npos || py == string :: npos )
		return 0;
		
	bool	sx = str [px-1] == '+';
	bool	sy = str [py-1] == '+';
	
	width   = atoi ( trim ( str.substr ( px + 1, py - px ) ).c_str () );
	height  = atoi ( trim ( str.substr ( py + 1 ) ).c_str () );
	
	return width > 0 && height > 0;
}

bool	loadRgbeRaw ( Data * data, byte * outBuf, int width, int height )
{
	int	count = height * width * 4;
	
	return data -> getBytes ( outBuf, count ) == count;
}

bool	loadRgbeRle ( Data * data, byte * outBuf, int width, int height )
{
	byte	px [4];
	byte * buffer = NULL;
	
	for ( int line = 0; line < height; line++ )
	{
		if ( data -> getBytes ( px, 4 ) != 4 )
			return false;
			
		if ( ( px [0] != 2 ) || ( px [1] != 2 ) || ( px [2] & 0x80 ) )
		{
			data -> seekCur ( -4 );
			
			return loadRgbeRaw ( data, outBuf, width, height );
		}
		
		int	len = ((int) px [2]) << 8 | px [3];
		
		if ( len != width )
			return false;
			
		if ( buffer == NULL )
			if ( ( buffer = (byte *) malloc ( 4*len ) ) == NULL )
				return false;
	    
	    byte * ptr = &buffer [0];
											
	    for ( int i = 0; i < 4; i++ )		// read each of the four channels for the scanline into the buffer 
		{
			byte * ptrEnd = &buffer [(i+1)*len];
		  
			while ( ptr < ptrEnd ) 
			{
			    if ( data -> getBytes ( px, 2 )  < 1 ) 
				{
					free ( buffer );
			    
					return false;
			    }
			
			    if ( px [0] > 128 ) 			// a run of the same value 
				{
					int	count = px [0] - 128;
				  
					if ( (count == 0) || (count > ptrEnd - ptr) ) 
					{
						free ( buffer );
						
						return false;
					}
				  
					memset ( ptr, px [1], count );
					ptr += count;

//					while ( count-- > 0 )
//						*ptr++ = buffer [1];
			    }
			    else 							// a non-run
				{
					int	count = px [0];
					
					if ( (count == 0) || (count > ptrEnd - ptr) ) 
					{
						free ( buffer );
						
						return false;
					}
					
					*ptr++ = px [1];
					
					if ( --count > 0 ) 
					{
						if ( data -> getBytes ( ptr, count ) < 1 )
						{
							free ( buffer );
							
							return false;
						}
						
						ptr += count;
					}
			    }
			}
	    }
												// copy byte data to output
	    for ( i = 0; i < width; i++, outBuf += 4 ) 
		{
		    outBuf [0] = buffer [i         ];
		    outBuf [1] = buffer [i +  width];
		    outBuf [2] = buffer [i +2*width];
		    outBuf [3] = buffer [i +3*width];
	    }
	}
  
	free ( buffer );
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////////

unsigned	loadRgbeAsRgba  ( Data * data, unsigned target, unsigned format, bool mipmap )
{
	unsigned	texture = 0;
	string		fmt;
	int			width, height;
	float		exposure, gamma;

	if ( !parseHdrHeader ( data, fmt, width, height, exposure, gamma ) )
		return 0;
		
	byte * buf = (byte *) malloc ( width * height * 4 );
	
	if ( buf == NULL )
		return 0;

	if ( !loadRgbeRle ( data, buf, width, height ) )
	{
		free ( buf );
		
		return 0;
	}
	
    glGenTextures   ( 1, &texture );
    glBindTexture   ( target, texture );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );                         // set 1-byte alignment

    glTexParameteri ( target, GL_TEXTURE_WRAP_S, GL_REPEAT );    // set texture to repeat mode
    glTexParameteri ( target, GL_TEXTURE_WRAP_T, GL_REPEAT );

    glTexImage2D ( target, 0, 4, width, height, 0, format, GL_UNSIGNED_BYTE, buf );

    if ( mipmap )
        glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    else
        glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		
    glTexParameteri ( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	free ( buf );
	
	return texture;
}

unsigned	loadRgbeAsFloat ( Data * data, unsigned target, unsigned format, bool mipmap )
{
	unsigned	texture = 0;
	string		fmt;
	int			width, height;
	float		exposure, gamma;

	if ( !parseHdrHeader ( data, fmt, width, height, exposure, gamma ) )
		return 0;
		
	int	   count = width * height;
	byte * buf   = (byte *) malloc ( count * 4 );
	
	if ( buf == NULL )
		return 0;

	if ( !loadRgbeRle ( data, buf, width, height ) )
	{
		free ( buf );
		
		return 0;
	}

	float * floatBuf = (float *) malloc ( 3 * sizeof ( float ) * count );
	
	if ( floatBuf == NULL )
	{
		free ( buf );
		
		return 0;
	}
	
	for ( int i = 0; i < count; i++ )
		rgbe2float ( floatBuf + i*3, buf + i*4 );
	
	free ( buf );
	
    glGenTextures   ( 1, &texture );
    glBindTexture   ( target, texture );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );                         // set 1-byte alignment

    glTexParameteri ( target, GL_TEXTURE_WRAP_S, GL_REPEAT );    // set texture to repeat mode
    glTexParameteri ( target, GL_TEXTURE_WRAP_T, GL_REPEAT );

    glTexImage2D ( target, 0, 4, width, height, 0, format, GL_FLOAT, floatBuf );

    if ( mipmap )
        glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    else
        glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		
    glTexParameteri ( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	free ( floatBuf );
	
	return texture;
}

